Option Explicit On 
Option Strict On
Option Compare Binary

Imports System
Imports System.Collections

' <remarks>
'   Ermittelt eine Lsung fr ein 0/1-Rucksackproblem.
' </remarks>
Public Class KnapsackSolver
    Private m_akobjObjects() As KnapsackObject
    Private m_dblKnapsackSize As Double

    ' Beste bisher gefundene Lsung.
    Private m_ablnMaxVec() As Boolean

    ' Maximale bisher gefundene Kosten.
    Private m_dblMaxCost As Double

    ' <summary>
    '   Lst das durch die in der Eigenschaft 
    '  <c>KnapsackSize</c> angegebene Grsse und die in der
    '   Eigenschaft <c>Objects</c> angebenen 
    '   <c>KnapsackObject</c>-Objekte spezifizierte Rucksackproblem
    '   und gibt die Lsung zurck.
    ' </summary>
    ' <returns>Lsung in Form eines 
    '    <c>KnapsackSolution</c>-Objekts.</returns>
    ' <exeption name="Exception">
    '   Wird geworfen, wenn die Eigenschaft <c>Objects</c> 
    '   nicht initialisiert wurde.
    ' </exception>
    Public Function Solve() As KnapsackSolution
        If Me.Objects Is Nothing Then
            Throw New Exception( _
              "Die Eigenschaft Objects muss ein gltiges Array" & _
              " von KnapsackObject-Objekten enthalten.")
        Else
            m_dblMaxCost = 0
            m_ablnMaxVec = New Boolean(Me.Objects.Length - 1) {}
            Dim v() As Boolean = New Boolean(Me.Objects.Length - 1) {}

            ' Enumeration durchfhren.
            Enumerate(-1, 0, 0, v)

            ' Da Arrays sehr unflexibel sind und wir auch 
            ' keine ArrayList verwenden wollten, mssen
            ' wir hier zuerst eine ArrayList der gewhlten 
            ' Elemente erstellen und diese dann in ein
            ' Array kopieren.
            Dim ks As KnapsackSolution = New KnapsackSolution()
            ks.TotalCost = m_dblMaxCost
            Dim i As Integer
            Dim al As New ArrayList()
            For i = 0 To m_akobjObjects.Length - 1
                If m_ablnMaxVec(i) Then
                    al.Add(m_akobjObjects(i))
                End If
            Next i
            Dim akobj() As KnapsackObject = _
              New KnapsackObject(al.Count - 1) {}
            For i = 0 To al.Count - 1
                akobj(i) = CType(al(i), KnapsackObject)
            Next i
            ks.SelectedObjects = akobj
            Return ks
        End If
    End Function

    ' <summary>
    '   Enumeriert ber alle gltigen Lsungen des 
    '   0/1-Ruchsackproblems. Danach findet sich
    '   die Lsung (gefllter Entscheidungsvektor) in 
    '   den entsprechenden globalen Variablen.
    ' </summary>
    Private Sub Enumerate( _
      ByVal z As Integer, _
      ByVal vcost As Double, _
      ByVal vweight As Double, _
      ByRef v As Boolean())
        Dim i As Integer
        If vweight <= Me.KnapsackSize Then
            If vcost > m_dblMaxCost Then
                m_dblMaxCost = vcost
                For i = 0 To v.Length - 1
                    m_ablnMaxVec(i) = v(i)
                Next i
            End If
            For i = z + 1 To Me.Objects.Length - 1
                v(i) = True
                Enumerate(i, vcost + Me.Objects(i).Cost, _
                  vweight + Me.Objects(i).Weight, v)
                v(i) = False
            Next i
        End If
    End Sub

    ' <summary>
    '   Gibt ein Array von <c>KnapsackObject</c>-Objekten 
    '   an oder gibt es zurck, die als Eingabedaten
    '   fr die Methode <c>Solve</c> verwendet werden sollen.
    ' </summary>
    ' <value>Objekte, die das Problem charakterisieren.</value>
    Public Property Objects() As KnapsackObject()
        Get
            Return m_akobjObjects
        End Get
        Set(ByVal Value() As KnapsackObject)
            m_akobjObjects = Value
        End Set
    End Property

    ' <summary>
    '   Gibt die Grsse des Rucksacks an oder gibt sie zurck.
    ' </summary>
    ' <value>Grsse des Rucksacks.</value>
    Public Property KnapsackSize() As Double
        Get
            Return m_dblKnapsackSize
        End Get
        Set(ByVal Value As Double)
            m_dblKnapsackSize = Value
        End Set
    End Property
End Class